﻿using BeatLyrics.Common.Models;
using BeatLyrics.Tool.Dialogs;
using BeatLyrics.Tool.Models;
using BeatLyrics.Tool.Utils;
using BrightIdeasSoftware;
using Microsoft.International.Converters.TraditionalChineseToSimplifiedConverter;
using System;
using System.Linq;
using System.Text.RegularExpressions;
using System.Windows.Forms;

namespace BeatLyrics.Tool.UserControls
{
    internal partial class TextArrangeUserControl : UserControl
    {
        private static readonly Regex RegRemoveRuby1 = new Regex("([（\\(]\\p{IsHiragana}+[\\)）])", RegexOptions.Compiled);
        private static readonly Regex RegRemoveRuby2 = new Regex("(<ruby=\\p{IsHiragana}+>)", RegexOptions.Compiled);
        private static readonly Regex RegRemoveRuby3 = new Regex("(</ruby>)", RegexOptions.Compiled);
        private static readonly Regex RegRubyConvert = new Regex("(<(.+?)=(\\p{IsHiragana}+)>)", RegexOptions.Compiled);

        private LyricFileExt _file;

        public TextArrangeUserControl()
        {
            InitializeComponent();
        }

        public void SetMediaInfo(int totalMilliseconds) => PlayPosTrackBar.Maximum = totalMilliseconds;

        public void SetLyricFile(LyricFileExt file)
        {
            _file = file;

            LeftLyricsObjectListView.SetObjects(file.Main);
            RightLyricsObjectListView.SetObjects(file.Subtitle);

            LeftTextButton.Enabled = !string.IsNullOrWhiteSpace(file.MainText);
            RightTextButton.Enabled = !string.IsNullOrWhiteSpace(file.SubtitleText);

            UpdateTimer.Start();
        }

        public void ClearData()
        {
            UpdateTimer.Stop();
            OggAudioPlayer.Unload();
        }

        private void MovePos(int ms)
        {
            foreach (LyricDetail item in LeftLyricsObjectListView.SelectedObjects)
            {
                item.TimeMs += ms;
            }

            LeftLyricsObjectListView.UpdateObjects(LeftLyricsObjectListView.SelectedObjects);

            foreach (LyricDetail item in RightLyricsObjectListView.SelectedObjects)
            {
                item.TimeMs += ms;
            }

            RightLyricsObjectListView.UpdateObjects(RightLyricsObjectListView.SelectedObjects);
        }

        private void SetPos(int ms)
        {
            foreach (LyricDetail item in LeftLyricsObjectListView.SelectedObjects)
            {
                item.TimeMs = ms;
            }

            LeftLyricsObjectListView.UpdateObjects(LeftLyricsObjectListView.SelectedObjects);

            foreach (LyricDetail item in RightLyricsObjectListView.SelectedObjects)
            {
                item.TimeMs = ms;
            }

            RightLyricsObjectListView.UpdateObjects(RightLyricsObjectListView.SelectedObjects);
        }

        private void SetEnd(int ms)
        {
            foreach (LyricDetail item in LeftLyricsObjectListView.SelectedObjects)
            {
                item.DurationMs = ms - item.TimeMs;
            }

            LeftLyricsObjectListView.UpdateObjects(LeftLyricsObjectListView.SelectedObjects);

            foreach (LyricDetail item in RightLyricsObjectListView.SelectedObjects)
            {
                item.DurationMs = ms - item.TimeMs;
            }

            RightLyricsObjectListView.UpdateObjects(RightLyricsObjectListView.SelectedObjects);
        }

        protected override bool ProcessCmdKey(ref Message msg, Keys keyData)
        {
            if (false == ActiveControl is TextBox && (keyData == Keys.Space || keyData == Keys.ProcessKey))
            {
                if (OggAudioPlayer.IsPlaying) OggAudioPlayer.Stop();
                else OggAudioPlayer.Play();
                return true;
            }
            return base.ProcessCmdKey(ref msg, keyData);
        }

        private void PlayButton_Click(object sender, EventArgs e)
        {
            if (OggAudioPlayer.IsPlaying) OggAudioPlayer.Stop();
            else OggAudioPlayer.Play();
        }

        private void UpdateTimer_Tick(object sender, EventArgs e)
        {
            var timeSpan = OggAudioPlayer.Position;
            PosLabel.Text = timeSpan.FormatToTotalMinuteAndSecondsAndMs();

            var milliseconds = (int)timeSpan.TotalMilliseconds;
            PlayPosTrackBar.Value = milliseconds > PlayPosTrackBar.Maximum
                ? PlayPosTrackBar.Maximum
                : milliseconds;

            if (_file == null)
            {
                CurrentLabel.Text = null;
                CurrentTranslatedLabel.Text = null;
            }
            else
            {
                CurrentLabel.Text = _file.Main?.FirstOrDefault(p => p.TimeMs <= milliseconds && p.TimeMs + p.DurationMs >= milliseconds)?.Text;
                CurrentTranslatedLabel.Text = _file.Subtitle?.FirstOrDefault(p => p.TimeMs <= milliseconds && p.TimeMs + p.DurationMs >= milliseconds)?.Text;
            }
        }

        private void PlayPosTrackBar_Scroll(object sender, EventArgs e)
        {
            OggAudioPlayer.Position = TimeSpan.FromMilliseconds(PlayPosTrackBar.Value);
        }

        private void PlayPosTrackBar_MouseDown(object sender, MouseEventArgs e)
        {
            if (OggAudioPlayer.IsPlaying) return;
            OggAudioPlayer.Play();
            PlayPosTrackBar.Tag = "Stop";
        }

        private void PlayPosTrackBar_MouseUp(object sender, MouseEventArgs e)
        {
            if (PlayPosTrackBar.Tag is string tag && tag == "Stop")
            {
                OggAudioPlayer.Stop();
                PlayPosTrackBar.Tag = null;
            }
        }

        private void LeftSelectAllButton_Click(object sender, EventArgs e) => LeftLyricsObjectListView.SelectAll();

        private void LeftDeleteButton_Click(object sender, EventArgs e)
        {
            foreach (LyricDetailExt item in LeftLyricsObjectListView.SelectedObjects)
            {
                _file.Main.Remove(item);
            }
            LeftLyricsObjectListView.SetObjects(_file.Main);
        }

        private void RemoveRubyButton_Click(object sender, EventArgs e)
        {
            foreach (LyricDetailExt item in LeftLyricsObjectListView.SelectedObjects)
            {
                item.Text = RegRemoveRuby1.Replace(item.Text, "");
                item.Text = RegRemoveRuby2.Replace(item.Text, "");
                item.Text = RegRemoveRuby3.Replace(item.Text, "");
                item.Text = RegRubyConvert.Replace(item.Text, match => match.Groups[2].Value);
                LeftLyricsObjectListView.UpdateObject(item);
            }
        }

        private void InsertRubyButton_Click(object sender, EventArgs e)
        {
            foreach (LyricDetailExt item in LeftLyricsObjectListView.SelectedObjects)
            {
                var arr = JapanesePhonetic.GetWords(item.Text);
                item.Text = string.Join("", arr.Select(p => p.ToString()));
                LeftLyricsObjectListView.UpdateObject(item);
            }
        }

        private void LeftTextButton_Click(object sender, EventArgs e)
        {
            new TextDialog { TextRichTextBox = { Text = _file.MainText } }.ShowDialog(this);
        }

        private void SetBeginButton_Click(object sender, EventArgs e) => SetPos(PlayPosTrackBar.Value);

        private void MoveForwardButton_Click(object sender, EventArgs e) => MovePos((int)MoveMsUpDown.Value);

        private void MoveBackwardButton_Click(object sender, EventArgs e) => MovePos(-(int)MoveMsUpDown.Value);

        private void SetEndButton_Click(object sender, EventArgs e) => SetEnd(PlayPosTrackBar.Value);

        private void RightTextButton_Click(object sender, EventArgs e)
        {
            new TextDialog { TextRichTextBox = { Text = _file.SubtitleText } }.ShowDialog(this);
        }

        private void TrimButton_Click(object sender, EventArgs e)
        {
            if (_file.Subtitle == null) return;

            foreach (var detail in _file.Subtitle)
            {
                detail.Text = detail.Text
                    .Trim()
                    .Trim('【', '】')
                    .Trim('(', ')')
                    .Trim('（', '）');
            }

            RightLyricsObjectListView.UpdateObjects(_file.Subtitle);
        }

        private void ZhCnButton_Click(object sender, EventArgs e)
        {
            foreach (LyricDetailExt item in RightLyricsObjectListView.SelectedObjects)
            {
                item.Text = ChineseConverter.Convert(item.Text, ChineseConversionDirection.TraditionalToSimplified);
            }

            RightLyricsObjectListView.UpdateObjects(RightLyricsObjectListView.SelectedObjects);
        }

        private void RightSelectAllButton_Click(object sender, EventArgs e) => RightLyricsObjectListView.SelectAll();

        private void RightDeleteButton_Click(object sender, EventArgs e)
        {
            foreach (LyricDetailExt item in RightLyricsObjectListView.SelectedObjects)
            {
                _file.Subtitle.Remove(item);
            }
            RightLyricsObjectListView.SetObjects(_file.Subtitle);
        }

        private void LyricsObjectListView_DoubleClick(object sender, EventArgs e)
        {
            var olv = (ObjectListView)sender;
            if (olv.SelectedObject != null)
                OggAudioPlayer.Position = TimeSpan.FromMilliseconds(((LyricDetail)olv.SelectedObject).TimeMs);
        }

        private void LyricsObjectListView_KeyDown(object sender, KeyEventArgs e)
        {
            var olv = (ObjectListView)sender;

            if (e.KeyCode == Keys.F2 && olv.SelectedObject != null)
            {
                var item = (LyricDetailExt)olv.SelectedObject;

                ContextDialog.PopTextBox(MousePosition, item.Text, (r, s) =>
                {
                    if (r == DialogResult.OK)
                    {
                        item.Text = s;
                        olv.UpdateObject(item);
                    }
                });
            }
        }
    }
}
